# -*- coding: utf-8 -*-
import json
import logging

from django.contrib.auth.models import Permission
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.db import transaction

from mptt.models import MPTTModel, TreeForeignKey

from .managers import MenuItemManager
from .exceptions import InvalidJson, UnsupportedJsonData, MissingJsonRequiredProp

logger = logging.getLogger(__name__)


class MenuItem(MPTTModel):
    label = models.CharField(_('label'),
                             max_length=50,
                             help_text=_('Display name on the website.'))
    slug = models.SlugField(
        _('slug'),
        unique=True,
        help_text=_('Unique identifier for the menu voice.'))
    parent = TreeForeignKey('self',
                            verbose_name=_('parent'),
                            on_delete=models.CASCADE,
                            null=True,
                            blank=True,
                            related_name='children')
    link = models.CharField(_('link'), max_length=255, blank=True, null=True)
    title = models.CharField(_('title'), max_length=255, blank=True, null=True)
    order = models.IntegerField(_('order'))
    enabled = models.BooleanField(_('enabled'), default=True)
    login_required = models.BooleanField(
        _('login required'),
        default=False,
        help_text=_(
            'If this is checked, only logged-in users will be able to view the item.'
        )  # noqa
    )
    permissions = models.ManyToManyField(
        Permission,
        verbose_name=_('permissions'),
        blank=True,
        help_text=_(
            'If empty, the menu item will be publicly visible, otherwise only users with at least one of the selected permissions could see it.'
        )  # noqa
    )

    objects = MenuItemManager()

    active = False
    with_active = False

    def __str__(self):
        return '{}'.format(self.label)

    class Meta:
        verbose_name = _('Menu item')
        verbose_name_plural = _('Menu items')

    class MPTTMeta:
        order_insertion_by = ('order', )

    @classmethod
    def from_json(cls, json_string):
        try:
            d = json.loads(json_string)
        except Exception as e:
            raise InvalidJson('Cannot parse provided json: %s' % str(e))

        if not isinstance(d, dict):
            raise UnsupportedJsonData('Provided json should be a dictionary containing a single root voice')

        with transaction.atomic():
            res = cls.create_item_from_dict(d)

        return res

    @classmethod
    def create_item_from_dict(cls, d, parent=None):
        if 'label' not in d or 'slug' not in d or 'order' not in d:
            raise MissingJsonRequiredProp('label, slug and order properties are mandatory')

        item = MenuItem.objects.create(
            parent=parent,
            label=d.get('label'),
            slug=d.get('slug'),
            order=d.get('order'),
            link=d.get('link', None),
            title=d.get('title', None),
            enabled=d.get('enabled', True),
            login_required=d.get('login_required', False),
        )

        for permission_code in d.get('permissions', []):
            item.permissions.add(Permission.objects.get(codename=permission_code))

        for child in d.get('children', []):
            cls.create_item_from_dict(child, item)

        return item

    def is_active(self, path):
        return self.link and self.link == path



# -*- coding: utf-8 -*-
import logging

from django import template
from django.contrib.auth.models import Permission
from django.db.models import Q

from ..models import MenuItem

register = template.Library()
logger = logging.getLogger(__name__)


def get_all_user_permissions_id_list(user):
    '''
    Returns a flat list of all user permissions ids
    '''
    # individual permissions
    perms = list(
        Permission.objects.filter(user=user).values_list('id', flat=True)
    )
    # permissions that the user has via a group
    group_perms = list(
        Permission.objects.filter(group__user=user).values_list('id', flat=True)
    )
    return perms + group_perms


def set_active_voice(path, items):
    '''
    Sets the active property to the active voice, and with_active to
    all its parents
    '''
    for item in items:
        if item.get('instance').is_active(path):
            item['active'] = True
            parent = item.get('parent')
            while parent is not None:
                parent['has_active'] = True
                parent = parent.get('parent')


def create_tree(context, root, parent=None):
    # get current user
    user = context.get('user')

    # if root is not enabled, do not show children
    if not root.enabled:
        return {}

    # if root visibility restrictions and user is not authenticated do not show children
    if not user.is_authenticated and (root.login_required or root.permissions.count()):  # noqa
        return {}

    if not user.is_authenticated:
        # unlogged user sees only public items
        items = root.children.enabled(login_required=False, permissions=None)
    elif user.is_superuser:
        # superuser sees all enabled items
        items = root.children.enabled()
    else:
        # logged in user which is not superuse should check for permissions
        permissions = context.get('lineup_permissions', None)
        if permissions is None:
            permissions = get_all_user_permissions_id_list(user)
            # put in context to avoid querying again at every iteration
            context['lineup_permissions'] = permissions

        items = root.children.enabled(
            Q(permissions__id__in=permissions) | Q(permissions=None)
        ).distinct()

    # parent needed to traverse upward for has-active functionality
    el = {'instance': root, 'parent': parent}
    el['children'] = [create_tree(context, child, el) for child in items]

    # search active voice
    if 'request' in context:
        path = context['request'].META['PATH_INFO']
        set_active_voice(path, el['children'])

    return el


@register.inclusion_tag('lineup/menu.html', takes_context=True)
def lineup_menu(context, item):
    '''
    Renders the whole tree if called recursively with different item param.
    - If item is a slug, the whole tree dictionary is created. and the first
    level items rendered.
    - If item is a dict, it just continues traversing rendering its children.
    '''
    if isinstance(item, str):
        try:
            root = MenuItem.objects.get(slug=item)
            tree = create_tree(context, root)
            items = tree.get('children', [])
            slug = item
            level = root.level
        except MenuItem.DoesNotExist:
            logger.error('Provided lineup menu slug %s not found' % item)
            return context
    elif isinstance(item, dict):
        items = item.get('children')
        slug = item.get('instance').slug
        level = item.get('instance').level
    else:
        items = []
        slug = None
        level = None

    context['items'] = items
    context['slug'] = slug
    context['level'] = level

    return context


@register.inclusion_tag('lineup/breadcrumbs.html', takes_context=True)
def lineup_breadcrumbs(context, slug):
    '''
    Renders menu breadcrumbs.
    '''
    if 'request' in context:
        path = context['request'].META['PATH_INFO']
        active_items = MenuItem.objects.enabled(link=path)

        found = False
        items = []
        for active_item in active_items:
            items.append(active_item)
            parent = active_item.parent
            while parent is not None and parent.enabled:
                items.insert(0, parent)
                parent = parent.parent
            if items.pop(0).slug == slug:
                found = True
                break

        if found:
            context['items'] = items
        else:
            context['items'] = []
    return context
